"""Update iBGP mesh workflow. Adds a new P router to the mesh of PE routers in the network."""

from typing import Any

from orchestrator.config.assignee import Assignee
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Label
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, begin, done, inputstep, step, workflow
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import ConfigDict, model_validator

from gso.products.product_blocks.router import RouterRole
from gso.products.product_types.router import Router
from gso.services import librenms_client, lso_client
from gso.services.lso_client import lso_interaction
from gso.services.subscriptions import get_trunks_that_terminate_on_router
from gso.utils.helpers import SNMPVersion, generate_inventory_for_active_routers
from gso.utils.types import TTNumber


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    """Show a confirmation window before running the workflow.

    Does not allow for user input, but does run a validation check. The workflow is only allowed to run, if the router
    already is connected by at least one IP trunk.
    """
    subscription = Router.from_subscription(subscription_id)

    class AddBGPSessionForm(FormPage):
        model_config = ConfigDict(title=f"Add {subscription.router.router_fqdn} to the iBGP mesh?")

        tt_number: TTNumber

        @model_validator(mode="before")
        def router_has_a_trunk(cls, data: Any) -> Any:
            terminating_trunks = get_trunks_that_terminate_on_router(
                subscription_id, SubscriptionLifecycle.PROVISIONING
            ) + get_trunks_that_terminate_on_router(subscription_id, SubscriptionLifecycle.ACTIVE)
            if len(terminating_trunks) == 0:
                msg = "Selected router does not terminate any available IP trunks."
                raise ValueError(msg)

            return data

    user_input = yield AddBGPSessionForm

    return user_input.model_dump()


@step("[DRY RUN] Add P router to iBGP mesh")
def add_p_to_mesh_dry(subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr) -> None:
    """Perform a dry run of adding the new P router to the PE router mesh."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Update iBGP mesh",
        "verb": "add_p_to_pe",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(RouterRole.PE),
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Add P router to iBGP mesh")
def add_p_to_mesh_real(subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr) -> None:
    """Add the P router to the mesh of PE routers."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Update iBGP mesh",
        "verb": "add_p_to_pe",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(RouterRole.PE),
        extra_vars=extra_vars,
    )


@step("[DRY RUN] Add all PE routers to P router iBGP group")
def add_all_pe_to_p_dry(subscription: dict[str, Any], callback_route: str) -> None:
    """Perform a dry run of adding the list of all PE routers to the new P router."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "pe_router_list": generate_inventory_for_active_routers(RouterRole.PE),
        "verb": "add_pe_to_p",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Add all PE routers to P router iBGP group")
def add_all_pe_to_p_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Add the list of all PE routers to the new P router."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "pe_router_list": generate_inventory_for_active_routers(RouterRole.PE),
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Update iBGP mesh",
        "verb": "add_pe_to_p",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("Verify iBGP session health")
def check_ibgp_session(subscription: Router, callback_route: str) -> None:
    """Run a playbook using the provisioning proxy, to check the health of the new iBGP session."""
    lso_client.execute_playbook(
        playbook_name="check_ibgp.yaml",
        callback_route=callback_route,
        inventory=subscription.router.router_fqdn,
        extra_vars={},
    )


@step("Add the router to LibreNMS")
def add_device_to_librenms(subscription: Router) -> State:
    """Add the router as a device to LibreNMS.

    If the device already exists, retrieve the device information instead of adding it again
    """
    client = librenms_client.LibreNMSClient()
    if not client.device_exists(subscription.router.router_fqdn):
        librenms_result = client.add_device(subscription.router.router_fqdn, SNMPVersion.V2C)
    else:
        librenms_result = client.get_device(subscription.router.router_fqdn)
    return {"librenms_device": librenms_result}


@inputstep("Prompt RADIUS insertion", assignee=Assignee.SYSTEM)
def prompt_insert_in_radius() -> FormGenerator:
    """Wait for confirmation from an operator that the router has been inserted in RADIUS."""

    class RADIUSPrompt(FormPage):
        model_config = ConfigDict(title="Please update RADIUS before continuing")

        info_label: Label = "Insert the router into RADIUS, and continue the workflow once this has been completed."

    yield RADIUSPrompt

    return {}


@inputstep("Prompt RADIUS login", assignee=Assignee.SYSTEM)
def prompt_radius_login() -> FormGenerator:
    """Wait for confirmation from an operator that the router can be logged into using RADIUS."""

    class RADIUSPrompt(FormPage):
        model_config = ConfigDict(title="Please check RADIUS before continuing")

        info_label: Label = "Log in to the router using RADIUS, and continue the workflow when this was successful."

    yield RADIUSPrompt

    return {}


@step("Update subscription model")
def update_subscription_model(subscription: Router) -> State:
    """Update the database model, such that it should not be reached via :term:`OOB` access anymore."""
    subscription.router.router_access_via_ts = False

    return {"subscription": subscription}


@workflow(
    "Update iBGP mesh",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def update_ibgp_mesh() -> StepList:
    """Update the iBGP mesh with a new P router.

    * Add the new P-router to all other PE-routers in the network, including a dry run.
    * Add all PE-routers to the P-router, including a dry run.
    * Verify that the iBGP session is up.
    * Add the new P-router to LibreNMS.
    * Update the subscription model.
    """
    return (
        begin
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> lso_interaction(add_p_to_mesh_dry)
        >> lso_interaction(add_p_to_mesh_real)
        >> lso_interaction(add_all_pe_to_p_dry)
        >> lso_interaction(add_all_pe_to_p_real)
        >> lso_interaction(check_ibgp_session)
        >> add_device_to_librenms
        >> prompt_insert_in_radius
        >> prompt_radius_login
        >> update_subscription_model
        >> resync
        >> done
    )
